/* Copyright 2015 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */

#include "../wireguard.h"
#include "noise.h"

bool noise_message_create_data(struct noise_message_data *dst, const u8 *src, size_t src_len, struct noise_peer *peer)
{
	bool ret = false;
	struct noise_session_keys *session;

	rcu_read_lock();
	session = rcu_dereference(peer->session);
	if (unlikely(!session || !session->sending))
		goto out;
	dst->header.type = NOISE_MESSAGE_DATA;
	dst->key_idx = cpu_to_le16(session->sending->index);
	if (unlikely(session->send_with_previous)) {
		if (likely(session->previous_sending))
			ret = noise_message_encrypt(dst, src, src_len, session->previous_sending);
		else
			ret = false;
	} else
		ret = noise_message_encrypt(dst, src, src_len, session->sending);
out:
	rcu_read_unlock();
	return ret;
}

/* Returns a strong reference to a peer */
struct wireguard_peer *noise_message_consume_data(u8 *dst, const struct noise_message_data *src, size_t src_len, struct wireguard_device *wg, bool *new_key)
{
	struct noise_session_keys *session;
	struct noise_peer *peer;
	struct wireguard_peer *wg_peer, *ret_peer = NULL;
	wg_peer = peer_lookup_by_index(wg, le16_to_cpu(src->key_idx));
	if (unlikely(!wg_peer))
		return NULL;
	peer = &wg_peer->noise;
	rcu_read_lock();
	session = rcu_dereference(peer->session);
	if (unlikely(!session || (!session->receiving && !session->previous_receiving))) {
		peer_put(wg_peer);
		goto out;
	}

	if (unlikely(!noise_message_decrypt(dst, src, src_len, session->receiving))) {
		if (likely(!noise_message_decrypt(dst, src, src_len, session->previous_receiving))) {
			peer_put(wg_peer);
			goto out;
		}
	} else if (unlikely(session->send_with_previous)) {
		session->send_with_previous = false;
		wmb();
		*new_key = true;
	}
	ret_peer = wg_peer;
out:
	rcu_read_unlock();
	return ret_peer;
}
